//
// -------------------------
// Autorun "Gateway" Program
// -------------------------
//	This is the program that is automatically run by the Windows 95
//	system whenever the primary GameFilm CD is inserted into a CD drive
//	(thanks to an entry in AUTORUN.INF, that is!)
//
//	This program detects whether or not the particular GameFilm title
//	has been installed on the user's system. If it has been installed,
//	then this program merely launches the game. If it has not yet been
//	installed, then this program allows the user the option of running
//	SETUP to install the game (or to cancel the gateway program).
//
//	Simple, but with pretty graphics!
//
//	Author: Jack Thornton
//	Additional graphics code: Jamie Fristrom
//	for CetaSoft / Level Five, Inc.
//	952 Postal Way, Suite 3
//	Vista, CA 92083
//	(619) 631-5855
//
//	Client: Atari Corporation
//	Project: GameFilm Engine
//	Date: 2/28/96
// --------------------------------------------------------------------------

#include <windows.h>
#include "splshdll.h"
#include "registry.h"

// ---------------------------------------------------- CONSTANT DECLARATIONS
// button graphic states
#define BTN_UNFOCUSED	0
#define BTN_FOCUSED	1
#define BTN_SELECTED	2

// button control ID base
#define ID_BTN		100

// ----------------------------------------------------- TYPES AND STRUCTURES
typedef struct {
    HWND hWnd;
    RECT btnRect;
    HBITMAP bitmap[3];
    RECT bitRects[3];
} ButtonInfo;

// --------------------------------------------------------- GLOBAL VARIABLES
char curDir[256];	    // current directory (should be CDROM)
char errorMessage[512];     // buffer for creating error messages
HINSTANCE hInstance;	    // instance of this application
BOOL bInstallationDetected; // flag: previous installation detected
HINSTANCE hResDLL;	    // handle to resource DLL
HWND hMainWindow;	    // main window handle
HBITMAP hMainBitmap;	    // handle to main splash-screen bitmap
RECT splashRect;	    // main splash-screen dimensions
ButtonInfo buttonTable[2];  // information about our owner-draw buttons
WNDPROC savedButtonProc;    // original button message handler

// DLL function entry points
BOOL (APIENTRY *pFetchBitmap)(UINT, LPBITMAPINFO *, LPVOID *, RECT *);
void (APIENTRY *pDiscardBitmap)(UINT);

// ----------------------------------------------------- PRE-INITIALIZED DATA
char APPPATH_KEY[] = {
  "Software\\Microsoft\\Windows\\CurrentVersion\\App Paths\\demogame.exe"
};

char APPPATH_VALUE[] = "demogame.exe";

char BIGPATH_KEY[] = {
  "Software\\Atari\\GameFilm\\CavesOfFear"
};

char TITLE_KEY[] = {"CavesOfFear"};

//char BIGPATH_CD_VALUE[] = "cmoo0";
char BIGPATH_CD_VALUE[] = "CMOO";

//char BIGPATH_HD_VALUE[] = "hmoo";
char BIGPATH_HD_VALUE[] = "HMOO";

char DLL_SRC_PATH[] = "GAMEFILM\\";

char GAME_WINDOW_CLASS[] = "L5";

char AUTORUN_WINDOW_CLASS[] = "GameFilmAutorun";

// --------------------------------------------------------- LOCAL PROTOTYPES
// internal routines
BOOL ValidateSystem();
BOOL InitAutorun();
BOOL InitMainWindow();
BOOL isGameAlreadyInstalled();
void CleanupAutorun();

// message handler subroutines
void mainWindowPaint(void);
void mainWindowCreate(HWND);
void mainWindowDrawItem(UINT, LPDRAWITEMSTRUCT);
void mainWindowCommand(WORD, WORD, DWORD);

// callbacks for windows & dialogs & such
LRESULT CALLBACK MessageHandler(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK ButtonSubclassProc(HWND, UINT, WPARAM, LPARAM);

// --------------------------------------------------------------------- CODE

//
// WinMain
//	The "starting point" for a Windows program.
//
int PASCAL
WinMain(HINSTANCE hInst, HINSTANCE, LPSTR, int)
{
MSG msg;
HWND otherAppWindow;

    // make handle to our instance globally visible
    hInstance = hInst;

    // do not run multiple instances of this
    otherAppWindow = FindWindow(AUTORUN_WINDOW_CLASS, NULL);
    if (otherAppWindow) {
	if (IsIconic(otherAppWindow))
	    ShowWindow(otherAppWindow, SW_RESTORE);
	SetForegroundWindow(otherAppWindow);
	return 0;
    }

    // perform system validation
    if (!ValidateSystem())
	return 0;

    // if game already installed, just launch it
    if (isGameAlreadyInstalled())
	return 0;

    // do other, process-related initializations
    if (!InitAutorun())
	return 0;

    // initialize the main window
    if (!InitMainWindow())
	return 0;

    // main message loop
    while (GetMessage(&msg, NULL, 0, 0)) {
	TranslateMessage(&msg);
	DispatchMessage(&msg);
    }

    // clean up
    CleanupAutorun();

    // sayonara, baby
    return (msg.wParam);
}

BOOL
ValidateSystem()
{
OSVERSIONINFO WindowsVersion;	// version information about operating system
//SYSTEM_INFO SysInfo;		  // system information

    // ensure that we are running on Windows 95
    WindowsVersion.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
    GetVersionEx(&WindowsVersion);
    if (WindowsVersion.dwPlatformId != VER_PLATFORM_WIN32_WINDOWS) {
	// if we are not running Win95, complain to user and bail
	MessageBeep(MB_ICONSTOP);
	MessageBox(NULL, "This program only runs under Windows 95",
	    "Cannot Execute Application", MB_ICONSTOP | MB_OK);
	return 0;
    }

//  // and on an Intel processor, at least a 80486
//  GetSystemInfo(&SysInfo);
//  if ((SysInfo.wProcessorArchitecture != PROCESSOR_ARCHITECTURE_INTEL) ||
//    (SysInfo.wProcessorLevel < 4)) {
//	// if we are not running an Intel 80486 or above, complain and bail
//	MessageBeep(MB_ICONSTOP);
//	MessageBox(NULL, "This program requires an Intel or compatable "
//	    "processor; minimum model 80486 or Pentium (or later)",
//	    "Cannot Execute Application", MB_ICONSTOP | MB_OK);
//	return 0;
//  }

    // passed the tests
    return TRUE;
}

BOOL
isGameAlreadyInstalled()
{
Registry reg;			// Windows "registry" object
char gameCommand[256];		// game application path, including command line
char gameExe[256];		// game executable fully-qualified-pathname
HWND runningGame;

    try {
	// See if a GameFilm is already running
	runningGame = FindWindow(GAME_WINDOW_CLASS, NULL);
	if (runningGame) {
	    if (IsIconic(runningGame))	// restore from icon/tray if minimized
		ShowWindow(runningGame, SW_RESTORE);
	    SetForegroundWindow(runningGame);	// activate game window
	    return TRUE;	// do not run this program if game already running
	}

	// look to see if the "bigfile" names are in the registry (and if they
	// actually exist in the specified places)
	if (GetFileAttributes(reg.readString(BIGPATH_HD_VALUE, BIGPATH_KEY,
	    HKEY_CURRENT_USER)) == 0xffffffff)
		return FALSE;
	if (GetFileAttributes(reg.readString(BIGPATH_CD_VALUE)) == 0xffffffff)
		return FALSE;

	// look in registry for execution path of the game
	reg.readStringIndirect(gameExe, 256, NULL, APPPATH_KEY,
	    HKEY_LOCAL_MACHINE);

	// build the command line
	wsprintf(gameCommand, "%s /n %s", gameExe, TITLE_KEY);

	// launch the game
	WinExec((LPSTR)gameCommand, SW_SHOW);

	// and do NOT continue with this program (somebody stop me...)
	return TRUE;
    }

    catch (Registry *) {
	// if there was a registry error, then the game is NOT installed
	return FALSE;
    }
}

BOOL
InitAutorun()
{
char filename[264];	    // name of file
HDC screenDC;		    // screen device context for creating HBITMAPs
LPBITMAPINFO header;	    // points to bitmap header
LPVOID lpBits;		    // points to bitmap data
int btn, state; 	    // loop counters

    // create a drive-designator string for the current path (CD-ROM)
    GetCurrentDirectory(256, curDir);
    curDir[3] = 0;			// just the drive designation & slash

    // load the bitmap resource DLL
    wsprintf(filename, "%s%sSPLASH.DLL", curDir, DLL_SRC_PATH);
    hResDLL = LoadLibrary(filename);
    if (!hResDLL) {
	wsprintf(errorMessage, "Unable to load or initialize module %s",
	    filename);
	MessageBeep(MB_ICONSTOP);
	MessageBox(NULL, errorMessage, "Error executing application",
	    MB_ICONSTOP | MB_OK);
	return FALSE;
    }

    // get the entry points for the DLL functions
    pFetchBitmap = (BOOL (APIENTRY *)(UINT, LPBITMAPINFO *, LPVOID *, RECT *))
	GetProcAddress(hResDLL, "_FetchBitmap@16");
    pDiscardBitmap = (void (APIENTRY *)(UINT))
	GetProcAddress(hResDLL, "_DiscardBitmap@4");
    if (!pFetchBitmap || !pDiscardBitmap) {
badmod: wsprintf(errorMessage, "Module %s is corrupted or incorrect", filename);
	MessageBeep(MB_ICONSTOP);
	MessageBox(NULL, errorMessage, "Error executing application",
	    MB_ICONSTOP | MB_OK);
	FreeLibrary(hResDLL);
	return FALSE;
    }

    // get a DC for reference for creating bitmap handles
    screenDC = GetDC(HWND_DESKTOP);

    // get the main bitmap
    if (!(*pFetchBitmap)(PIC_SPLASH_SCREEN, &header, &lpBits, &splashRect))
	goto badmod;
    hMainBitmap = CreateDIBitmap(
	screenDC,
	(BITMAPINFOHEADER *)header,
	CBM_INIT,
	lpBits,
	header,
	(((BITMAPINFOHEADER *)header)->biClrUsed == 256) ?
	   DIB_PAL_COLORS : DIB_RGB_COLORS
	);

    // fetch the button bitmaps
    for (btn = 0; btn < 2; btn++) {
	for (state = 0; state < 3; state++) {
	    // fetch the bitmap information for this button state
	    if (!(*pFetchBitmap)(PIC_BTNINSTALL_UNFOCUSED + (btn * 3) + state,
	      &header, &lpBits, &buttonTable[btn].bitRects[state]))
		goto badmod;

	    // create a DDB from the DIB and save the handle
	    buttonTable[btn].bitmap[state] = CreateDIBitmap(
		screenDC,
		(BITMAPINFOHEADER *)header,
		CBM_INIT,
		lpBits,
		header,
		(((BITMAPINFOHEADER *)header)->biClrUsed == 256) ?
		   DIB_PAL_COLORS : DIB_RGB_COLORS
	    );

	    // the button rectangle will be the union of all state rects
	    UnionRect(&buttonTable[btn].btnRect, &buttonTable[btn].btnRect,
		&buttonTable[btn].bitRects[state]);
	}

	// now change all of the individual bitmap rects so that they
	// are relative to the overall button window rect
	for (state = 0; state < 2; state++) {
	    buttonTable[btn].bitRects[state].left -= buttonTable[btn].btnRect.left;
	    buttonTable[btn].bitRects[state].right -= buttonTable[btn].btnRect.left;
	    buttonTable[btn].bitRects[state].top -= buttonTable[btn].btnRect.top;
	    buttonTable[btn].bitRects[state].bottom -= buttonTable[btn].btnRect.top;
	}
    }

    // thanks for the dc!
    ReleaseDC(screenDC, HWND_DESKTOP);

    // okay
    return TRUE;
}

BOOL
InitMainWindow()
{
WNDCLASSEX wclass;	    // class definition for our window
RECT rWorkArea; 	    // screen/workarea dimensions
int x, y, width, height;    // dimensions for main window

    // register window class
    wclass.cbSize = sizeof(WNDCLASSEX);
    wclass.style = CS_BYTEALIGNCLIENT | CS_BYTEALIGNWINDOW | CS_HREDRAW |
	CS_VREDRAW | CS_OWNDC;
    wclass.lpfnWndProc = (WNDPROC) MessageHandler;
    wclass.cbClsExtra = 0;
    wclass.cbWndExtra = 0;
    wclass.hInstance = hInstance;
    wclass.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(32000));
    wclass.hCursor = LoadCursor(NULL, IDC_ARROW);
    wclass.hbrBackground = (HBRUSH) (COLOR_APPWORKSPACE + 1);
    wclass.lpszMenuName = NULL;
    wclass.lpszClassName = (LPSTR)AUTORUN_WINDOW_CLASS;
    wclass.hIconSm = (HICON)LoadImage(hInstance, MAKEINTRESOURCE(32000),
	IMAGE_ICON, 16, 16, 0);
    if (!RegisterClassEx(&wclass))
	return FALSE;

    // calculate necessary dimensions for our window
    if (!SystemParametersInfo(SPI_GETWORKAREA, sizeof(RECT), &rWorkArea, 0)) {
	rWorkArea.left = rWorkArea.top = 0;
	rWorkArea.right = GetSystemMetrics(SM_CXSCREEN);
	rWorkArea.bottom = GetSystemMetrics(SM_CYSCREEN);
    }
    width = (splashRect.right - splashRect.left) + GetSystemMetrics(SM_CXFRAME);
    height = (splashRect.bottom - splashRect.top) + GetSystemMetrics(SM_CYFRAME) +
	GetSystemMetrics(SM_CYCAPTION);
    x = ((rWorkArea.right - rWorkArea.left) - width) / 2;
    y = ((rWorkArea.bottom - rWorkArea.top) - height) / 2;

    // create window
    hMainWindow = CreateWindow(
	(LPSTR)AUTORUN_WINDOW_CLASS,
	"GameFilm CD-ROM",
	WS_OVERLAPPED | WS_CAPTION | WS_THICKFRAME |
	  WS_SYSMENU | WS_MINIMIZEBOX | WS_MAXIMIZEBOX |
	  WS_CLIPCHILDREN | WS_CLIPSIBLINGS | WS_VISIBLE,
	x,
	y,
	width,
	height,
	HWND_DESKTOP,
	NULL,
	hInstance,
	NULL
	);

    // success?
    return (hMainWindow != NULL);
}

void
CleanupAutorun()
{
int i;		// loop counter

    // free the main bitmap
    if (hMainBitmap)
	DeleteObject(hMainBitmap);

    // robustness check
    if (hResDLL) {
	// free up the bitmaps
	if (pDiscardBitmap)
	    for (i = 0; i < 16; i++)
		(*pDiscardBitmap)(i);

	// free the bitmap resource DLL
	FreeLibrary(hResDLL);
    }
}

LRESULT CALLBACK
MessageHandler(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg) {
	case WM_CREATE:
	    mainWindowCreate(hWnd);
	    break;

	case WM_PAINT:
	    mainWindowPaint();
	    break;

	case WM_DRAWITEM:
	    mainWindowDrawItem((UINT) wParam, (LPDRAWITEMSTRUCT) lParam);
	    break;

	case WM_COMMAND:
	    mainWindowCommand(LOWORD(wParam), HIWORD(wParam), lParam);
	    break;

	case WM_DESTROY:
	    PostQuitMessage(0);
	    break;
    }

    // default message processing
    return DefWindowProc(hWnd, msg, wParam, lParam);
}

void
mainWindowCreate(HWND hWnd)
{
int btn;	    // loop counter

    // create the buttons
    for (btn = 0; btn < 2; btn++) {
	// create the button
	buttonTable[btn].hWnd = CreateWindow(
	    "BUTTON",
	    "",
	    WS_CHILD | WS_VISIBLE | BS_OWNERDRAW,
	    buttonTable[btn].btnRect.left,
	    buttonTable[btn].btnRect.top,
	    buttonTable[btn].btnRect.right - buttonTable[btn].btnRect.left,
	    buttonTable[btn].btnRect.bottom - buttonTable[btn].btnRect.top,
	    hWnd,
	    (void*) (ID_BTN + btn),
	    hInstance,
	    NULL
	    );

	// subclass the button
	savedButtonProc = (WNDPROC) SetWindowLong(buttonTable[btn].hWnd,
	    GWL_WNDPROC, (LONG)ButtonSubclassProc);
    }
}

void
mainWindowPaint()
{
PAINTSTRUCT ps; 	    // information structure for "painting" window
HDC hDC;		    // device context for this window
HDC memDC;		    // a "memory" device context

    // Begin paint response (and get our DC)
    hDC = BeginPaint(hMainWindow, &ps);

    // Create our memory DC/bitmap buffer
    memDC = CreateCompatibleDC(hDC);
    SaveDC(memDC);

    // Select the bitmap object into the memory DC
    SelectObject(memDC, hMainBitmap);

    // Blit splash screen bitmap to window
    BitBlt(hDC,
	   0,
	   0,
	   splashRect.right - splashRect.left,
	   splashRect.bottom - splashRect.top,
	   memDC,
	   0,
	   0,
	   SRCCOPY);

    // clean up
    RestoreDC(memDC, -1);
    DeleteDC(memDC);

    // done painting
    EndPaint(hMainWindow, &ps);
}

void
mainWindowDrawItem(UINT idctl, LPDRAWITEMSTRUCT lpdis)
{
HDC memDC;		// memory device context
ButtonInfo *button;	// points to entry of corresponding button in table
int stateIndex; 	// index into button rectangle/shape subtables

    // determine which button we are talking about
    button = &buttonTable[idctl - ID_BTN];

    // create a memory DC for blitting
    memDC = CreateCompatibleDC(lpdis->hDC);
    SaveDC(memDC);

    // select proper button table index based on button state
    if (lpdis->itemState & ODS_SELECTED)  // if selected
	stateIndex = BTN_SELECTED;
    else
	stateIndex = (lpdis->itemState & ODS_FOCUS) ?
	    BTN_FOCUSED : BTN_UNFOCUSED;

    // first, draw background to ensure erasure of larger button images
    SelectObject(memDC, hMainBitmap);
    BitBlt(lpdis->hDC, 0, 0, button->btnRect.right - button->btnRect.left,
	button->btnRect.bottom - button->btnRect.left, memDC,
	button->btnRect.left, button->btnRect.top, SRCCOPY);

    // next, draw correct button image for current state
    SelectObject(memDC, button->bitmap[stateIndex]);
    BitBlt(lpdis->hDC, button->bitRects[stateIndex].left,
	button->bitRects[stateIndex].top,
	button->bitRects[stateIndex].right - button->bitRects[stateIndex].left,
	button->bitRects[stateIndex].bottom - button->bitRects[stateIndex].top,
	memDC, 0, 0, SRCCOPY);

    // clean up
    RestoreDC(memDC, -1);
    DeleteDC(memDC);
}

void
mainWindowCommand(WORD controlID, WORD event, DWORD parameter)
{
    // were one of our buttons clicked?
    if (event == BN_CLICKED) {
	// which one?
	switch (controlID) {
	    case ID_BTN + 0:
		// launch the installer
		lstrcat(curDir, "SETUP.EXE");
		WinExec(curDir, SW_SHOW);
		break;

	    case ID_BTN + 1:
		// nothing - just quit program
		break;
	}

	// in any case, this program goes away
	DestroyWindow(hMainWindow);
    }
}

LRESULT CALLBACK
ButtonSubclassProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
POINT xy;
RECT clientArea;
LRESULT retval;
WPARAM msgParm1;

    switch (msg) {
	case WM_MOUSEMOVE:
	    // mouse already been captured by us
	    if (GetCapture()) {
		xy.x = LOWORD(lParam);
		xy.y = HIWORD(lParam);
		GetClientRect(hWnd, &clientArea);
		if (!PtInRect(&clientArea, xy)) {
		    // mouse left
		    ReleaseCapture();
		    SetFocus(GetParent(hWnd));
		}
	    } else {
		// capture mouse so we'll know when it's left us:
		SetCapture(hWnd);
		SetFocus(hWnd);
		//PlaySound("GOTFOCUS", hInstance, SND_RESOURCE);
	    }
	    break;

	case WM_LBUTTONDOWN:
	    // allow default processing to occur
	    retval = (*savedButtonProc)(hWnd, msg, wParam, lParam);

	    // play sound effect for clicking on button
	    //PlaySound("CLICKED", hInstance, SND_RESOURCE);

	    // send "button clicked" message to parent window
	    msgParm1 = MAKELONG(GetWindowLong(hWnd, GWL_ID), BN_CLICKED);
	    SendMessage(GetParent(hWnd), WM_COMMAND, msgParm1, (LPARAM)hWnd);

	    // return result from default processing
	    return retval;

	case WM_LBUTTONUP:
	    return 0;
    }

    // default message processing
    return (*savedButtonProc)(hWnd, msg, wParam, lParam);
}

