/********************************************************************\

	Wave Module
	by Rudy Mathieu

\********************************************************************/

#include "wave2.h"
#include "dbgnew.h"
#include "frsrc.h"
#include <assert.h>
#include <stdexcpt.h>
#define DBUFFERSIZE 		(0x400 * 16)   	/* 16K buffer */
#define DATACHUNKSIZE		(0x400 * 4)		/* File Chunk 4K */
#define WAVE_CLOCK_DELAY	32				/* Timer Interval */
#define WAVE_RESOLUTION		50				/* Accuracy */

static LONG soundTimers=0;
static void CALLBACK periodicWaveChecker(UINT uID, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2)
{
   CWave *wave = (CWave *)dwUser;
   InterlockedIncrement(&soundTimers);
   wave->Check();
   InterlockedDecrement(&soundTimers);
}

CWave::~CWave()
{
  if (state == CW_PLAYING)
  {
    Stop();
  }
  if (state == CW_OPEN)
  {
    Close();
  }
  if (state != CW_READY)
  {
    throw logic_error("bad CWave state");
  }
  assert(state == CW_READY);
  if (reader)
  {
    delete reader;
  }
}

CWave::CWave()
{
  bytesToWrite = 0;
  end_pos = INVALID_VALUE;
  timeEvent = 0;
  *(int*)&status = 0;
  sample = NULL;
  dsound = NULL;
  reader = DBG_NEW(MMIOReader);
  state = CW_READY;
}

void CWave::Close(void)
{
  assert(state==CW_OPEN || state==CW_READY || state==CW_PLAYING);
  reader->Close();
  CloseSound();
  state = CW_READY;
}

void CWave::SetReader(CWaveReader* nr) 
{
 if (reader) delete reader; reader = nr; 
}

void CWave::CloseSound(void)
{
  assert(state==CW_OPEN || state==CW_READY || state==CW_PLAYING);
  if (sample)
  {
    delete[] sample;
    sample = NULL;
  }
  if (dsound)
  {
    delete dsound;
    dsound = NULL;
  }
  state = CW_READY;
}

DWORD CWave::SetupSound(DWORD size,WAVEFORMATEX *fmt)
{
   CloseSound();
   dataChunk = DATACHUNKSIZE;
   if (size > DBUFFERSIZE)
     size = DBUFFERSIZE;

   dsound = DBG_NEW(DSoundBuffer(fmt,size));
   if (dsound==NULL)
     return 0;

   sample = DBG_NEW_N(BYTE,dataChunk);
   if (sample==NULL)
   {
     CloseSound();
     return 0;
   }
   status.sampleReady = FALSE;
   return size;
}

BOOL CWave::Open(const char *filename)
{
  WAVEFORMATEX fmt;
  if (state == CW_PLAYING)
  {
    Stop();
  }
  if (state == CW_OPEN)
  {
    Close();
  }
  assert(state == CW_READY);
  
  if (reader->Open(filename))
  {
    if (reader->GetFormat(fmt))
    {
      soundSize = SetupSound(reader->GetLength(), &fmt);
		if (soundSize)
		{
    	  Reset();
        state = CW_OPEN;
        return TRUE;
      }
    }
  }
  Close();
  return FALSE;
}

BOOL CWave::Open(int index)
{
  WAVEFORMATEX fmt;
  if (state == CW_PLAYING)
  {
    Stop();
  }
  if (state == CW_OPEN)
  {
    Close();
  }
  assert(state == CW_READY);

  if (reader->Open(index))
  {
    if (reader->GetFormat(fmt))
    {
      soundSize = SetupSound(reader->GetLength(), &fmt);
		if (soundSize)
		{
        reader->Rewind();
        status.eofReached = FALSE;
        dsound->SetPlayPosition(0);
        state = CW_OPEN;
        return TRUE;
      }
    }
  }
  Close();
  return FALSE;
}

void CWave::Reset(void)
{ 
  assert(state==CW_OPEN || state==CW_READY || state==CW_PLAYING);
  Stop(); 
  reader->Rewind();
  status.eofReached = FALSE;
  dsound->SetPlayPosition(0);
}

BOOL CWave::Play(int loopcount)
{
  assert(state==CW_OPEN || state==CW_READY || state==CW_PLAYING);
  Stop();
  if (state == CW_OPEN)
  {
    status.donePlaying =FALSE;
    loop = loopcount;
    Check();
    timeEvent = timeSetEvent(WAVE_CLOCK_DELAY,WAVE_RESOLUTION,(LPTIMECALLBACK)periodicWaveChecker,(DWORD)(this),TIME_PERIODIC);
    dsound->Play(DSBPLAY_LOOPING);
    state = CW_PLAYING;
  }
  return TRUE;
}

BOOL CWave::Stop(void)
{ 
  assert(state==CW_OPEN || state==CW_READY || state==CW_PLAYING);
  if (state == CW_PLAYING)
  {
    status.donePlaying = TRUE;
    while (soundTimers) Sleep(0);
    if (timeEvent)
    {
      timeKillEvent(timeEvent);
      timeEvent = 0;
    }
    dsound->Stop(); 
    state = CW_OPEN;
  }
  return TRUE;
}

BOOL CWave::StopUnguarded(void)
{ 
  assert(state==CW_OPEN || state==CW_READY || state==CW_PLAYING);
  if (state == CW_PLAYING)
  {
    status.donePlaying = TRUE;
    if (timeEvent)
    {
      timeKillEvent(timeEvent);
      timeEvent = 0;
    }
    dsound->Stop(); 
    state = CW_OPEN;
  }
  return TRUE;
}

BOOL CWave::Looping(void)
{
  loop--;
  if (loop < 0)
    loop = -1;

  if (loop)
  {
    reader->Rewind();
    status.eofReached = 0;
    return TRUE;
  }
  return FALSE;
}

LONG CWave::ReadSoundBite(void)
{
  assert(state==CW_OPEN || state==CW_READY || state==CW_PLAYING);
  if (!(status.sampleReady || status.eofReached))
  {
    LONG amount;
    if (!reader->Read((BYTE*)sample, dataChunk, amount))
    {
      if (Looping()==FALSE)
      {
        status.eofReached = TRUE;
      }
      return 0;
    }
    status.sampleReady = TRUE;
    return amount;
  }
  return 0;
}
//------------------------------------------------
void CWave::Check(void)
{
  if (status.donePlaying)
    return;

  // New: try to have the
  // sound stop on it's own
  // when there are no more samples.
  // (Apparently SoundStream is not finished)
  if (status.eofReached && loop==0)
  {
    status.donePlaying = TRUE;
    dsound->Stop();
    return;
  }
  if (!status.sampleReady)
  {
    /* Get data from disk */
    int b = ReadSoundBite();
    if (b) bytesToWrite = b;
  }
  if (status.sampleReady)
  {
    end_pos = INVALID_VALUE;
    if (dsound->FillSoundBuffer(sample,bytesToWrite)!=INVALID_VALUE)
    {
      status.sampleReady = 0;
    }
  } 
  else 
  {
    if (end_pos == INVALID_VALUE)
    {
      end_pos = dsound->FillSoundBuffer(sample,dataChunk);
    }
    else 
    {
      if (dsound->PlayPointerInside(end_pos,dataChunk)==TRUE)
      {
        Reset();
      }
    }
  }
}

BOOL MMIOReader::Open(const char* filename)
{
  h = mmioOpen((LPTSTR)filename,NULL,MMIO_READ);
  if (h)
  {
    return TRUE;
  }
  return FALSE;
}

BOOL MMIOReader::Close()
{
  if (h)
  {
    mmioClose(h,0);
  }
  h = 0;
  return TRUE;
}

LONG MMIOReader::GetLength()
{
  return length;
}
BOOL MMIOReader::GetFormat(WAVEFORMATEX& fmt)
{
  MMCKINFO parent;
  MMCKINFO child;
  parent.fccType = mmioFOURCC('W','A','V','E');
  if (mmioDescend(h,&parent,NULL,MMIO_FINDRIFF)==MMSYSERR_NOERROR)
  {
    child.ckid = mmioFOURCC('f','m','t',' ');
    if (mmioDescend(h,&child,&parent,MMIO_FINDCHUNK)==MMSYSERR_NOERROR)
    {
	   if (mmioRead(h, (HPSTR) &fmt, child.cksize) == (LONG)child.cksize)
  	   {
   	  mmioAscend(h,&child,0);
	  	  child.ckid = mmioFOURCC('d','a','t','a');
   	  if (mmioDescend(h,&child,&parent,MMIO_FINDCHUNK)==MMSYSERR_NOERROR)
		  {
		    topOfData = mmioSeek(h,0,SEEK_CUR);
          length = child.cksize;
          return TRUE;
        }
      }
    }
  }
  return FALSE;
}

BOOL MMIOReader::Rewind()
{
  mmioSeek(h,topOfData,SEEK_SET);
  return TRUE;
}

BOOL MMIOReader::Read(BYTE* dest, LONG bytesToRead, LONG& bytesRead)
{
  LONG amount = mmioRead(h, (HPSTR) dest, bytesToRead);
  if (amount == 0)
  {
    return FALSE;
  }
  bytesRead = amount;
  return TRUE;
}

BOOL ResourceReader::Open(const char* filename)
{
  HRSRC r = FindResource(0, filename, "WAVE");
  HGLOBAL gh = LoadResource(0, r);
  BYTE* ptr = (BYTE*)LockResource(gh);
  int fmtPos = 20; // offset of content of format chunk.
  int nextPos = *(LONG*)(ptr+fmtPos-4) // size of format chunk
               + fmtPos // plus it's position
               ;
  while (*(long*)(ptr+nextPos) != mmioFOURCC('d','a','t','a'))
  {
    nextPos = nextPos + 4 + 4 + *(long*)(ptr+nextPos+4); // skip content of useless chunk.
  }
  int dataPos = nextPos + 4 + 4;
  dataLength = *(LONG*)(ptr+dataPos-4);
  fmt = *(WAVEFORMATEX*)(ptr+fmtPos);
  //data = new BYTE[dataLength];
  data = DBG_NEW_N(BYTE, dataLength);
  memcpy(data, ptr+dataPos, dataLength);
  FreeResource(gh);
  position = 0;
  return TRUE;
}
BOOL ResourceReader::GetFormat(WAVEFORMATEX& fmt_)
{
  fmt_ = fmt;
  return TRUE;
}

BOOL ResourceReader::Rewind()
{
  position = 0;
  return TRUE;
}

BOOL ResourceReader::Close()
{
  if (data)
  {
  delete [] data;
  data = 0;
  dataLength = 0;
  }
  return TRUE;
}

BOOL ResourceReader::Read(BYTE* dest, LONG bytesToRead, LONG& bytesRead)
{
  if (position+bytesToRead <= dataLength)
  {
    memcpy(dest, data+position, bytesToRead);
    bytesRead = bytesToRead;
    position += bytesRead;
    return TRUE;
  }
  return FALSE;
}
//------------------------------------------------
FRSRC* BigFileWaveReader::bigfile = 0;
//------------------------------------------------
BOOL BigFileWaveReader::Close()
{
  bigfile->CloseFile(handle);
  dataLength = 0;
  return TRUE;
}

//------------------------------------------------
BOOL BigFileWaveReader::GetFormat(WAVEFORMATEX& fmt_)
{
  fmt_ = fmt;
  return TRUE;
}

//------------------------------------------------
BOOL BigFileWaveReader::Open(const char* filename)
{
  handle = bigfile->OpenFile(filename);
  if (handle != -1)
  {
    if (bigfile->Read(handle, &fmt, sizeof(fmt)) == sizeof(fmt))
    {
      if (bigfile->Read(handle, &dataLength, sizeof(dataLength))==sizeof(dataLength))
      {
        dataStart = sizeof(fmt) + sizeof(dataLength);
        position = dataStart;
        return TRUE;
      }
    }
  }
  return FALSE;
}

//------------------------------------------------
BOOL BigFileWaveReader::Open(int index)
{
  handle = bigfile->OpenFileEntry(index);
  if (handle != -1)
  {
    if (bigfile->Read(handle, &fmt, sizeof(fmt)) == sizeof(fmt))
    {
      if (bigfile->Read(handle, &dataLength, sizeof(dataLength))==sizeof(dataLength))
      {
        dataStart = sizeof(fmt) + sizeof(dataLength);
        position = dataStart;
        return TRUE;
      }
    }
  }
  return FALSE;
}

//------------------------------------------------
BOOL BigFileWaveReader::Read(BYTE* dest, LONG bytesToRead, LONG& bytesRead)
{
  if (position+bytesToRead <= dataStart+dataLength)
  {
    long r = bigfile->Read(handle, dest, bytesToRead);
    if (r >= 0 && r <= bytesToRead)
    {
      bytesRead = r;
      position += bytesRead;
      return TRUE;
    }
  }
  return FALSE;
}

//------------------------------------------------
BOOL BigFileWaveReader::Rewind()
{
  position = dataStart;
  long r = bigfile->Seek(handle, dataStart, FRSRC_BEGIN);
  return r == dataStart;
}

/*---------------------------------------------------------------------------*\
							 Sound Bite Methods
\*---------------------------------------------------------------------------*/

DWORD SoundBite::SetupSound(DWORD size,WAVEFORMATEX *fmt)
{
  CloseSound();
  dataChunk = size;
  //dsound = new DSoundBuffer(fmt,size);
  dsound = DBG_NEW(DSoundBuffer(fmt,size));
  if (dsound==NULL)
    return 0;

  //sample = new BYTE[dataChunk];
  sample = DBG_NEW_N(BYTE, dataChunk);
  if (sample==NULL)
  {
    CloseSound();
    return 0;
  }
  status.sampleReady = FALSE;
  if (ReadSoundBite() < (LONG)size)
  {
    CloseSound();
    return 0;
  }
  if (dsound->FillSoundBuffer(sample,dataChunk)==INVALID_VALUE)
  {
    CloseSound();
    return 0;
  }

  delete[] sample; /* We don't need this anymore */
  sample=NULL;
  status.sampleReady = 1;
  return size;
}

void SoundBite::Reset(void)
{
  Stop(); 
  dsound->SetPlayPosition(0);
}

BOOL SoundBite::Play(int loopcount)
{
  Stop();
  loop = loopcount + 1;
  if (Looping()==FALSE)
    return FALSE;
  status.donePlaying = FALSE;
  timeEvent = timeSetEvent(WAVE_CLOCK_DELAY,WAVE_RESOLUTION,(LPTIMECALLBACK)periodicWaveChecker,(DWORD)(this),TIME_PERIODIC);
  state = CW_PLAYING;
  return TRUE;
}

BOOL SoundBite::Looping(void)
{
  if (loop==0) return FALSE;
  loop--;
  if (loop < 0)
    loop = -1;

  if (loop)
  {
    dsound->SetPlayPosition(0);
    dsound->Play(0);
    return TRUE;
  }
  return FALSE;
}

void SoundBite::Check(void)
{
  if (status.donePlaying)
    return;

  if (dsound->DonePlaying())
  {
    if (Looping()==FALSE)
    {
      StopUnguarded();
    }
  }
}
